import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:dio/adapter.dart';
import 'package:dio/dio.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart';
import 'package:http_parser/http_parser.dart';
import 'package:intl/intl.dart';
import 'package:mime/mime.dart';
import 'package:snosey_flutter_package/Enums.dart';
import 'package:snosey_flutter_package/api/ApiResponse.dart';

class ApiController {
  static late String baseUrl;
  static String? appSecret;
  static String? apiKey;
  static String? userAgent;
  static late String serverErrorMessage;
  static String? refreshTokenUrl;
  static Map<String, String> staticParams = {};
  static String _cookies = "";
  final String apiUrl;
  final RequestTypeEnum requestType;
  static Map<String, String> staticHeaders = {};
  static ByteData? _pemCer;
  static String? _keyCer;

  Map<RequestTypeEnum, String> requestTypeMap = {
    RequestTypeEnum.PUT: "PUT",
    RequestTypeEnum.POST: "POST",
    RequestTypeEnum.GET: "GET",
    RequestTypeEnum.DELETE: "DELETE"
  };

  ApiController(this.apiUrl, this.requestType);

  static init({
    required String baseUrl,
    required String serverErrorMessage,
  }) {
    ApiController.baseUrl = baseUrl;
    ApiController.serverErrorMessage = serverErrorMessage;
  }

  static _addCookies(String key, String value) {
    _cookies += "$key=$value;";
  }

  Future<ApiResponse> sendRequest(
      {Map<String, dynamic> body = const {},
      Map<String, File> files = const {},
      bool isMultiPart = false,
      Map<String, String> header = const {},
      Map<String, String> queryParameters = const {}}) async {
    staticParams.addAll(queryParameters);
    if (_cookies.isNotEmpty) staticHeaders["Cookie"] = _cookies;
    staticHeaders.addAll(header);
    var requestData;
    if (isMultiPart) {
      var requestData = FormData();
      if (files.isNotEmpty) {
        requestData.files.addAll(files.map((fileKey, file) {
          var mime = lookupMimeType(file.path);
          var fileByte = file.readAsBytesSync();
          var fileName = "${file.path.split('/').last}";
          var contentType = MediaType(mime!.split('/')[0], mime.split('/')[1]);
          return MapEntry(
              fileKey,
              MultipartFile.fromBytes(
                fileByte,
                filename: fileName,
                contentType: contentType,
              ));
        }).entries);
      }
      if (body.isNotEmpty) {
        body.forEach((key, value) {
          requestData.fields.add(fillMultiPart(key, value, "")!);
        });
      }
    } else {
      requestData = body;
    }
    var api = Dio(
      BaseOptions(
        baseUrl: baseUrl,
        contentType: 'application/json; charset=UTF-8',
        responseType: ResponseType.json,
        headers: staticHeaders,
        method: requestType.toString().replaceAll("RequestTypeEnum.", ""),
      ),
    );

    if (_pemCer != null) {
      (api.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate =
          (client) {
        SecurityContext sc = SecurityContext();
        sc.setClientAuthoritiesBytes(
          _pemCer!.buffer.asUint8List(),
        //  password: "_keyCer",
        );
        HttpClient httpClient = HttpClient(context: sc);
        return httpClient;
      };
    }

    api.interceptors.add(
      LogInterceptor(
        responseBody: true,
        error: true,
        request: true,
        requestBody: true,
        requestHeader: true,
        responseHeader: true,
      ),
    );
    var response = await api
        .request(
          apiUrl,
          data: requestData,
          queryParameters: staticParams,
        )
        .onError((error, stackTrace) => throw serverErrorMessage);
    if (response.statusCode == HttpStatus.ok) {
      if (response.headers.value("authorization") != null) {
        Auth._setBearerToken(
            token: response.headers.value("authorization")!,
            expiration: response.headers.value("expires")!);
        response.headers["set-cookie"]!.forEach((cookie) {
          Auth._setRefreshToken(cookie);
        });
      }

      var apiResponse = ApiResponse();
      apiResponse.setStatus(jsonDecode(response.headers.value("x-status")!));
      if (response.headers.value("x-pagination") != null) {
        apiResponse.pagination.hasPagination = true;
        apiResponse.pagination
            .set(jsonDecode(response.headers.value("x-pagination")!));
      } else {
        apiResponse.pagination.hasPagination = false;
      }
      if (apiResponse.success) {
        apiResponse.objectResponse = response.data;
        return apiResponse;
      } else
        throw utf8.decode(base64.decode(apiResponse.errorMessage));
    } else {
      throw serverErrorMessage;
    }
  }

  MapEntry<String, String>? fillMultiPart(
      String key, dynamic value, String root) {
    if (value is Map) {
      for (var object in value.entries) {
        return fillMultiPart(object.key, object.value, key + ".");
      }
    } else if (value is List) {
      for (int i = 0; i < value.length; i++)
        return fillMultiPart("$key[$i]", value[i], root);
    } else {
      return MapEntry("$root$key", value.toString());
    }
  }
}

class Auth {
  static var expirationFormat = DateFormat("E, dd MMM yyy HH:mm:ss 'GMT'");
  static Timer? reNewExpirationTimer;

  static init({
    required String appSecret,
    required UserAgentEnum userAgent,
    required String apiKey,
    required String refreshTokenUrl,
    required ByteData pemCer,
    required String keyCer,
  }) {
    ApiController.refreshTokenUrl = refreshTokenUrl;
    ApiController._addCookies("Secret", appSecret);
    ApiController._addCookies("Api-Key", apiKey);
    ApiController.staticHeaders["User-Agent"] =
        userAgent.toString().replaceFirst("UserAgentEnum.", "");
    ApiController._pemCer = pemCer;
    ApiController._keyCer = keyCer
        .replaceAll("-----BEGIN PRIVATE KEY-----", "")
        .replaceAll("\n", " ")
        .replaceAll("-----END PRIVATE KEY-----", "");
  }

  static _setBearerToken({
    required String token,
    required String expiration,
  }) {
    ApiController.staticHeaders["Authorization"] = token;
    var expirationMin = DateTime.now()
            .difference(expirationFormat.parse(expiration))
            .inMinutes -
        5;
    print("renew in $expirationMin min");
    reNewExpirationTimer = Timer(Duration(minutes: expirationMin), () {
      reNewToken();
    });
  }

  static Future<bool> reNewToken() async {
    final storage = new FlutterSecureStorage();
    String? expires = await storage.read(key: "expires");
    if (expires != null &&
        expirationFormat.parse(expires).isAfter(DateTime.now())) {
      String? token = await storage.read(key: "refreshToken");
      var response = await ApiController(
              ApiController.refreshTokenUrl!, RequestTypeEnum.POST)
          .sendRequest(body: {"Token": token!});
      return response.success;
    } else {
      return false;
    }
  }

  static clearToken() {
    if (reNewExpirationTimer != null) reNewExpirationTimer!.cancel();
    if (ApiController.staticHeaders.containsKey("Authorization"))
      ApiController.staticHeaders.remove("Authorization");
    final storage = new FlutterSecureStorage();
    storage.deleteAll();
  }

  static _setRefreshToken(String cookies) {
    final storage = new FlutterSecureStorage();
    cookies.split(";").forEach((element) {
      var key = element.split("=").first;
      var value = element.split("=").last;
      if (key == "refreshToken" || key == "expires") {
        storage.write(key: key, value: value);
      }
    });
  }
}
